"""A simple module to wrap the built in open and close functions so that it prints
OPENING and CLOSING every time a file is opened/closed, and can print out the list
of all open files if requested. This module can be used to track down file handle leaks.
"""

import builtin
openfiles = set()

oldfile = builtin.file
oldopen = builtin.open

class newfile(oldfile):
    """Class to wrap the open file function"""
    def __init__(self, *args):
        self.name = args[0]
        print("*** OPENING {0} ***".format(str(self.name)))
        oldfile.__init__(self, *args)
        openfiles.add(self)

    def close(self):
        """wraps the close function"""
        print("*** CLOSING {0} ***".format(str(self.name)))
        oldfile.close(self)
        openfiles.remove(self)

def newopen(*args):
    """The function to wrap the file/open commands, so that we can log the filenames being
    opened"""
    return newfile(*args)

def listOpenFiles():
    """Returns the list of open files"""
    return openfiles #

def printOpenFiles():
    """Prints the list of open files"""
    print ("*** {0} OPEN FILES: [{1}]".format((len(openfiles), ", ".join(f.name for f in openfiles))))

builtin.file = newfile
builtin.open = newopen



import nose

class Testing(object):
    def __init__(self):
        return

    def inRange(self,res, irlow,irhigh=None):
        """Returns true if res between the ranges of ir
        
        Arguments:
        res -- variable being tested, float, int, numpy array
        irlow -- lower bound
        irhigh -- upper bound
        
        Output:
        Boolean variable (possibly array) of whether irlow<=res<irhigh
        
        """
        try:
            return (res>=irlow[0])&(res<irlow[1])
        except:
            return (res>=irlow)&(res<irhigh)
    
    def withinFractionalError(self, a, b, e=None):
        """Returns true if a=b+-e
        
        Arguments:
        a -- result being tested
        b -- target value
        e -- Fraction error bars to test within against
        
        Output:
        boolean result of a=b+-e
        
        """
        if e is None:
            e=1e-4
        if (a+b) is not 0:
            return (abs((a-b)*2.0/(a+b)) < e)
        else :
            return (abs((a-b)) < e)

    def withinAbsoluteValue(self, a, b, e):
        """Returns true if a=b+-e
        
        Arguments:
        a -- result being tested
        b -- target value
        e -- error bars to test within against
        
        Output:
        boolean result of a=b+-e
        
        """
        return self.inRange(a,b-e,b+e)
    
    def debug(self, funcname):
        """Prints a nice message 
        
        Arguments:
        functname -- name of test function that called debug
        
        """
        print("Testing...{0}".format(funcname))

    
    def test_inRange(self):
        self.debug("inRange")
        assert(self.inRange(1.0,0.0,2.0))
        assert(self.inRange(1.1,1.09,1.11))
        assert(not self.inRange(0.0,1.09,1.11))

    def test_withinAV(self):
        self.debug("withinAV")
        assert(self.withinAbsoluteValue(0.0,0.0,1e-4))
        assert(self.withinAbsoluteValue(0.0,1e-5,1e-4))
        assert(not self.withinAbsoluteValue(0.0,1.0,0.99))
        
    def test_withinFR(self):
        self.debug("withinFR")
        assert(self.withinFractionalError(1.0,1.0,1e-4))
        assert(self.withinFractionalError(1.0,1.01,0.1))
        assert(self.withinFractionalError(1.0,1.01,0.01))
        assert(not self.withinFractionalError(1.0,1.01,0.009))
        assert(not self.withinFractionalError(0.0,1.0,1e-3))
        assert(not self.withinFractionalError(0.0,1.0,1e-3))

    def test_testing(self):
        assert True
        assert not False
        self.test_inRange()
        self.test_withinAV()
        self.test_withinFR()


if __name__=="__main__":
    t=Testing()
    t.test_testing()



import numpy


def smooth(data, weights, return_weights=False, allow_edge_effects=False):
    """Smooths 1-dimensional data using numpy.convolve based on the weights
    
    :param data: The input array
    :type data: numpy array
    
    :param weights:   The weights array, length M (not dependent on N)
    :type weights: numpy array
    
    :param return_weights: whether to include the weights and the output in a tuple
    :type return_weights: logical
    
    :return: A smoothed array of length N, possibly in a tuple with the weights
    :rtype: numpy array
    """
    result = numpy.convolve(data, weights, 'same')
    if not allow_edge_effects:
        #edge correct
        try:
            lw = len(weights)
        except:
           lw = weights.size 
        nw = (lw-1)/2
        s_start = slice(0,lw,1)
        s_end = slice(-lw,None,1)
        result[0:lw] = data[0:lw]
        result[-lw:] = data[-lw:]
    
    if return_weights is not None:
        if return_weights:
            return (result, weights)
        
    return result    

def oddbox(funcname, boxsize, quiet=False, *args, **kwargs):
    """Ensure that the boxsize for smoothing is centered
    
    Makes a smoothing box/window be 2M+1 long to put the result in the centre box
    
    :param funcname: Function name, just for the message 
    :type funcname: string
    
    :param boxsize: 2M|2M+1 that is corrected to be 2M+1
    :type boxsize: integer
    
    :param quiet: Determines if the warning is NOT printed, default is false.
    :type quiet: logical
    
    :return: Odd number.
    :rtype: integer
    """
    if boxsize % 2 == 0 and not quiet:
        print("""boxsize should be odd ({0} smoothing),
currently {1}""".format(funcname, boxsize))
        boxsize+=1

    return boxsize

def scipy_window_smooth(data, boxsize=9,window_name='triangle',options=None, *args, **kwargs):
    """
    Use a scipy defined window to smooth. Options include 
    'triangle','boxcar','gaussian','blackman','parzen','hamming'.
    
    All but the 'gaussian' option takes a single option of the window size. 'gaussian'
    also takes the width.
    
    :param data: The input array
    :type data: numpy array
    
    :param boxsize:  total width of the smoothing window
    :type boxsize: integer
    
    :param window_name: scipy defined name of the smoothing function
    
    :return: Results from smooth, which is either the smoothed data, or 
      smoothed data and weights used in the smooothing
    :rtype: numpy array
    """
    
    import scipy.signal
    if window_name == 'gaussian':
        input = (window_name, options)
    else:
        input = window_name
    
    weights = scipy.signal.get_window(input, boxsize)
    weights = weights / numpy.sum(weights)
    return smooth(data, weights, *args, **kwargs)
    
def box(data, boxsize=9, *args, **kwargs):
    """Smoothing with a uniform distribution
    
    Generates an array of length N filled with values of 1/N, then calls smooth with this window
    
    :param data: The input array
    :type data: numpy array
    
    :param boxsize:  total width of the smoothing window
    :type boxsize: integer
    
    :return: Results from smooth, which is either the smoothed data, or
      smoothed data + weights used in the smooothing
    :rtype: numpy array
    """
    boxsize = oddbox("box", boxsize, *args, **kwargs)
    weights = numpy.ones(boxsize) * (1./boxsize)
    weights = weights/numpy.sum(weights)
    return smooth(data, weights, *args, **kwargs)

def triangle(data, boxsize=9, *args, **kwargs):
    """Smoothing with a triangular distribution
    
    Generates an array of length N filled with values in a symmetric 
    triangular distribution, then calls smooth with this window
    
    :param data: The input array
    :type data: numpy array
    
    :param boxsize:  total width of the smoothing window
    :type boxsize: integer
    
    :return: Results from smooth, which is either the smoothed data, or smoothed data 
      and weights used in the smooothing
    :rtype: numpy array
    """
    boxsize = oddbox("box", boxsize, *args, **kwargs)
    weights = numpy.arange(boxsize)
    weights = boxsize/2. - numpy.abs(weights - numpy.mean(weights))
    weights = weights / numpy.sum(weights)
    return smooth(data, weights, *args, **kwargs)
    
def gaussian(data, boxsize=9, nsigma=3, *args, **kwargs):
    """Smoothing with a gaussian distribution of size boxsize that covers nsigma widths
    
    Generates an array of length N filled with values of 1/N, then calls smooth with this window
    
    Arguments:
    
    :param data: The input array
    :type data: numpy
    
    :param boxsize:  total width of the smoothing window
    :type boxsize: integer
    
    :param nsigma: number of sigmas (std devs) to contain in the len(boxsize) elements. 
    :type nsigma: float
    
    Larger values of sigma give a tighter distribution in the centre and more delta-function 
    like behaviour, low sigma value gives a smoother distribution with more uniform like
    behaviour.
    
    :return: Results from smooth, which is either the smoothed data, or smoothed data 
      and weights used in the smooothing.
    :rtype: numpy array
    """
    boxsize = oddbox("box", boxsize, *args, **kwargs)
    sigma = boxsize/ ((nsigma+1.0e-10)*numpy.sqrt(2*numpy.log(2)))
    weights = numpy.arange(boxsize)
    weights = (weights - numpy.mean(weights))
    weights = numpy.exp(-(weights/sigma)**2)
    weights = weights/numpy.sum(weights)

    return smooth(data, weights, *args, **kwargs)
    



import netCDF4
from collections import OrderedDict as odict
import six

def copy_attributes(ncin, ncout,exclude=None, include=None):
    """Copy the global or variable attributes from the input to the output"""
    att_dict = odict()
    for attribute_name in ncin.ncattrs():
        if include is not None and attribute_name not in include:
            continue #if include is defined, and this attribute is not there
        if exclude is not None and attribute_name in exclude:
            continue #if exclude is defined, and this attribute is there
        att_dict[attribute_name]  = ncin.getncattr(attribute_name)
    ncout.setncatts(att_dict)

def copy_dimensions(ncin, ncout, exclude=None, include=None):
    """Copy the file dimensions from input to output"""
    for dimname, dimval in six.iteritems(ncin.dimensions):
        if include is not None and dimname not in include:
            continue #if include is defined, and this attribute is not there
        if exclude is not None and dimname in exclude:
            continue #if exclude is defined, and this attribute is there

        size = None if dimval.isunlimited() else len(dimval)
        ncout.createDimension(dimname, size=size)

def copy_variable_data(ncin, ncout, include=None, exclude=None):
    "Copy the variable information, but not the data, from input to output"
    
    for varname, variable in six.iteritems(ncin.variables):
        if include is not None and varname not in include:
            continue #if include is defined, and this attribute is not there
        if exclude is not None and varname in exclude:
            continue #if exclude is defined, and this attribute is there

        datatype = variable.dtype
        dimensions = variable.dimensions
        fill_value = None
        if "missing_value" in variable.ncattrs():
            fill_value = variable.getncattr("missing_value")
        if "_FillValue" in variable.ncattrs():
            fill_value = variable.getncattr("_FillValue")
        outvariable = ncout.createVariable(varname, datatype, dimensions, fill_value=fill_value)
        
        #get the attributes
        copy_attributes(variable, outvariable, exclude=["missing_value","_FillValue"])
        
def copy_variables(ncin, ncout,include=None, exclude=None):
    copy_variable_data(ncin, ncout, include=include, exclude=exclude)
    for varname, variable in six.iteritems(ncin.variables):
        if include is not None and varname not in include:
            continue #if include is defined, and this attribute is not there
        if exclude is not None and varname in exclude:
            continue #if exclude is defined, and this attribute is there

        ncout.variables[varname][:] = variable[:]
    
def copy_netcdf(input_filename, output_filename, *args, **kwargs):
    ncin = netCDF4.Dataset(input_filename)
    ncout = netCDF4.Dataset(output_filename, 'w',*args, **kwargs)
    
    copy_attributes(ncin, ncout)
    copy_dimensions(ncin, ncout)
    copy_variables(ncin, ncout)

    ncin.close()
    ncout.close()
    
def ensemble_define_variables(ncin, ncout, dimension_name="mode"):
    #create new dimension
    ncout.createDimension(dimension_name, 2)
    ensemble_variables = []
    for varname, variable in six.iteritems(ncin[0].variables):
        datatype = variable.dtype
        dimensions = variable.dimensions
        if len(dimensions) == 1 and varname in dimensions:
            #this is an axis dimension, skip the calculation.
            pass
        else:
            #not an axis dimension, save the name for later, add the dimension
            ensemble_variables.append(varname)
            dimensions = list(dimensions)
            dimensions.insert(1,dimension_name)
        fill_value = None
        if "missing_value" in variable.ncattrs():
            fill_value = variable.getncattr("missing_value")
        if "_FillValue" in variable.ncattrs():
            fill_value = variable.getncattr("_FillValue")
        outvariable = ncout.createVariable(varname, datatype, dimensions, fill_value=fill_value)
        #get the attributes
        copy_attributes(variable, outvariable, exclude=["missing_value","_FillValue"])
    
    iDenom = 1./len(ncin)
    for varname in ensemble_variables:
        data = ncin[0].variables[varname]
        if data.dtype == "S1":
            ncout.variables[varname][:,0] = data[:]
            continue
        #else
        m, v = data[:] * iDenom, data[:] * data[:] * iDenom
        if len(ncin) > 1:
            for nc in ncin[1:]:
                data = nc.variables[varname]
                m,v = m+data[:]*iDenom, v+data[:]*data[:]*iDenom
        v=v-m*m
        ncout.variables[varname][:,0] = m
        ncout.variables[varname][:,1] = v
        
def ensemble_mean_variance(input_filenames, output_filename, *args, **kwargs):
    ncin = [netCDF4.Dataset(f) for f in input_filenames]
    ncout = netCDF4.Dataset(output_filename, 'w',*args, **kwargs)
    
    copy_attributes(ncin[0], ncout)
    copy_dimensions(ncin[0], ncout)
    #add ensemble mean/variance dimension    
    ensemble_define_variables(ncin, ncout)
    
    for n in ncin:
        n.close()
    ncout.close()



import numpy as np
import dwell.planck as planck
import dwell.constants as const
import nose.tools
from dwell.testing import Testing

class test_planck(object):
    def __init__(self):
        """Initialize variables used to test planck library
        
        Initialized the variables needed to run the tests, these include
        solar temperature, wavelength range, solar_radius/orbit ratio,
        wavenumber range, acutal results from the integration, actual
        results form the series approximation.
        
        """
        self.test = Testing()
        self.solar_temp = 5750.
        self.wavelength = np.linspace(0.01,500,100000)
        self.scale = 0.004652**2 # (solar_rad/1.au)^2
        self.wavenumber = np.linspace(1,100000,1000)
        self.actual =  [1341.41, 1341.42]
        self.actual_cfst4 = [1341.41, 1341.42]
        
    def debug(self, expected, result,funcname):
        self.test.debug(funcname)

        print("expected between {0} and {1}".format(
            expected[0], expected[1]))
        print("{0} = {1}".format(funcname,result))

    def test_planck_wavelength(self):
        """Testing Planck wavelength with solar flux
        
        Calculates Planck function at a series of wavelengths, integrates to 
        get the total solar flux at the Earth, units of W/m^2
        
        """
        b = planck.planck_wavelength(self.wavelength,
                                     self.solar_temp)*self.scale
        tb= np.sum(b)*(self.wavelength[1]-self.wavelength[0])* np.pi
        self.debug(self.actual,tb,"planck_wavelength")
        assert (self.test.inRange(tb, self.actual))
    
    def test_planck_wavenumber(self):
        """Testing Planck wavenumber with solar flux
        
        Calculates Planck function at a series of wavenumbers, integrates to 
        get the total solar flux at the Earth, units of W/m^2.
        
        """
        b = planck.planck_wavenumber(self.wavenumber,
        self.solar_temp)*self.scale
        tb= np.sum(b)*(self.wavenumber[1]-self.wavenumber[0])* np.pi
        self.debug(self.actual,tb,"planck_wavenumber")
        assert (self.test.inRange(tb, self.actual))


    def test_st4(self):
        """Test sigma*T^4
        
        Tests st4 function with solar temperature, scaled by radius and orbit to 
        give solar flux at the Earth. Units of W/m^2
        
        """
        tb = planck.st4(self.solar_temp)*self.scale
        self.debug(self.actual,tb,"planck_st4")
        assert (self.test.inRange(tb, self.actual))
        
    def test_cfst4_wavenumber(self):
        """Testing wavenumber integrated planck
        
        Tests cfst4 with solar temperature, scaled by radius and orbit to 
        give solar flux at the Earth. Units of W/m^2
        
        """
        
        tb = planck.cfst4_wavenumber(self.wavenumber[0],
                                     self.wavenumber[-1],
                                     self.solar_temp)*self.scale
        self.debug(self.actual_cfst4,tb,"planck_cfst4_wavenumber")
        assert (self.test.inRange(tb, self.actual_cfst4))

        
    def test_cfst4_wavelength(self):
        """Testing wavelength integrated planck
        
        Tests cfst4 with solar temperature, scaled by radius and orbit to 
        give solar flux at the Earth. Units of W/m^2
        
        """
        tb = planck.cfst4_wavelength(self.wavelength[0],
                                     self.wavelength[-1],
                                     self.solar_temp)*self.scale
        self.debug(self.actual_cfst4,tb,"planck_cfst4_wavelength")
        assert (self.test.inRange(tb, self.actual_cfst4))

    def test_cfst4_a(self):
        """Testing cfst4 with finite limits
        
        Tests cfst4 with solar temperature, over finite wavelength limits
        form 1.0 to 100 micron, scaled by radius and orbit to 
        give solar flux at the Earth. Units of W/m^2
        
        """
        wl = 0.1
        wh = 1000.
        t = 300.
        error=1e-6
        ta = planck.cfst4_wavelength(wl,wh,t)
        tb = planck.cfst4_wavenumber(10000/wh,10000/wl,t)
        print("Comparing cfst4_wavelength and cfst4_wavenumber")
        print("cfst4_wavelength={0}".format(ta))
        print("cfst4_wavenumber={0}".format(tb))
        assert self.test.inRange((ta-tb),[-error,error])
        
    def allt(self):
        """Run all tests"""
        self.test_planck_wavelength()
        self.test_planck_wavenumber()
        self.test_st4()
        self.test_cfst4_wavenumber()
        self.test_cfst4_wavelength()
        self.test_cfst4_a()


if __name__=="__main__":
    c = test_planck()
    c.allt()
    



"""Module to generate deprecation warnings for functions that are no longer used.
To use this package, import it (`from dwell.RAD.deprecated import deprecated`) and
wrap a deprecated function with @deprecated.
Optional arguments include a nicer message (for wrapping class methods, say) in `message`,
a `replacement` function, and a requirement to `use_replacement` instead of the called
function _with the same arguments_
"""
#This module is BSD licensed

license="BSD"
import warnings
import dwell.rad.terminal as t
warnings.simplefilter("always")

class _deprecated(object):
    """deprecated class"""
    def __init__(self, function, message=None, replacement=None, use_replacement=False):
        """Initialize deprecation message. This initialization takes the `function`
        to be deprecated, an optional nice `message`, a possible `replacement` function,
        and an instruction to `use_replacement` instad of the old function."""

        self.function = function
        self.message = message
        self.replacement = replacement
        self.use_replacement = use_replacement
        self.__name__ = self.function.__name__
        self.__doc__ = self.function.__doc__
        self.__dict__.update(self.function.__dict__)

    def __call__(self,*args,**kwargs):
        """Wrapper to function() so that you can call this class as if it were
        a function"""
        #build the message
        if self.message is None:
            message = "{0} is deprecated.".format(self.function.__name__)
        else:
            message = self.message
        #notify of a replacement
        if self.replacement is not None:
                message += " use {0} instead".format(self.replacement.__name__)
        #warn the user, in magenta.
        warnings.warn(t.warning(message), DeprecationWarning, stacklevel=2)

        #call the function, or replacement
        try:
            if self.use_replacement:
                return self.replacement(*args, **kwargs)
            else:
                return self.function(*args, **kwargs)
        except TypeError as e:
            if self.use_replacement:
                return self.replacement()
            else:
                return self.function()

def deprecated(function=None, message=None, replacement=None, use_replacement=False):
    """Wrapper function that deals with the possibility of keyword arguments that
    results in the default argument (the function) being None. In this case, a closure
    function is generated (wrapper) that stores the keyword arguments and returns
    a new wrapper function (which then wraps the function being deprecated, but with
    hidden arguments)."""
    if function:
        return _deprecated(function)
    else:
        def wrapper(function):
            return _deprecated(function, message, replacement, use_replacement)
        return wrapper



from dwell.testing import Testing
import dwell.smoothing as sm
import numpy


class test_smoothing(object):
    def __init__(self):
        """Initialize variables used to test smoothing library
        """
        self.test = Testing()
    
    def test_oddbox(self):
        assert sm.oddbox("dummy",11) == 11
        assert sm.oddbox("dummy",10) == 11

    def test_box(self):
        data = numpy.ones(100)
        r,w=sm.box(data, 9,return_weights=True)
        assert numpy.all(self.test.withinAbsoluteValue(r-data,0.0,1e-10))
        assert numpy.all(self.test.withinAbsoluteValue(w,1.0/9.0,1e-10))
    
        r,w=sm.box(data, 8,return_weights=True)
        assert numpy.all(self.test.withinAbsoluteValue(r-data,0.0,1e-10))
        assert numpy.all(self.test.withinAbsoluteValue(w,1.0/9.0,1e-10))

    def test_regress_generate(self):
        """regression test of window generation"""
        
        boxsize=11
        #box
        data = numpy.ones(boxsize)
        
        regression_tests = dict(box=dict(
                                    function=sm.box,
                                    target=numpy.array([ 0.09090909,  0.09090909,  0.09090909,  
                                                         0.09090909,  0.09090909,  0.09090909,  
                                                         0.09090909,  0.09090909,  0.09090909,  
                                                         0.09090909,  0.09090909]),
                                    args=boxsize),
                                triangle=dict(
                                    function=sm.triangle,
                                    target=numpy.array([0.01639344,  0.04918033,  0.08196721,
                                                        0.1147541 ,  0.14754098,  0.18032787,  
                                                        0.14754098,  0.1147541 ,  0.08196721,  
                                                        0.04918033,  0.01639344]),
                                    args=boxsize),
                                gaussian=dict(
                                    function=sm.gaussian,
                                    target=numpy.array([ 0.01392149,  0.03521418,  0.07247478,
                                                         0.12136501,  0.16536266,  0.18332377,
                                                         0.16536266,  0.12136501,  0.07247478,
                                                         0.03521418,  0.01392149]),
                                    args=boxsize)
                                )
        scipy_regression_tests = dict(
                                    triangle=dict(
                                    function=sm.scipy_window_smooth,
                                    window_name='triangle',
                                    target=numpy.array([ 0.02777778,  0.05555556,  0.08333333,  
                                                         0.11111111,  0.13888889,  0.16666667,  
                                                         0.13888889,  0.11111111,  0.08333333,  
                                                         0.05555556,  0.02777778]),
                                    args=boxsize),
                                    boxcar=dict(
                                    function=sm.scipy_window_smooth,
                                    window_name='boxcar',
                                    target=numpy.array([ 0.09090909,  0.09090909,  0.09090909,  
                                                         0.09090909,  0.09090909,  0.09090909,  
                                                         0.09090909,  0.09090909,  0.09090909,  
                                                         0.09090909,  0.09090909]),
                                    args=boxsize),
                                    gaussian=dict(
                                    function=sm.scipy_window_smooth,
                                    window_name='gaussian',
                                    target=numpy.array([ 0.03548293,  0.05850147,  0.08630959,  
                                                         0.1139453 ,  0.13461047,  0.14230046, 
                                                         0.13461047,  0.1139453 ,  0.08630959,  
                                                         0.05850147,  0.03548293]),
                                    args=boxsize),

                                    blackman=dict(
                                    function=sm.scipy_window_smooth,
                                    window_name='blackman',
                                    target=numpy.array([ -3.30423519e-18,   9.57449104e-03,   4.78024151e-02,
                                                          1.21377890e-01,   2.02197585e-01,   2.38095238e-01,
                                                          2.02197585e-01,   1.21377890e-01,   4.78024151e-02,
                                                          9.57449104e-03,  -3.30423519e-18]),
                                    args=boxsize),

                                    parzen=dict(
                                    function=sm.scipy_window_smooth,
                                    window_name='parzen',
                                    target=numpy.array([ 0.00036423,  0.00983427,  0.04552905,  
                                                         0.12001457,  0.20305955,  0.24239665,  
                                                         0.20305955,  0.12001457,  0.04552905,  
                                                         0.00983427,  0.00036423 ]),
                                    args=boxsize),

                                    hamming=dict(
                                    function=sm.scipy_window_smooth,
                                    window_name='hamming',
                                    target=numpy.array([0.01459854,  0.03062996,  0.07260076,  
                                                        0.12447953,  0.16645033,  0.18248175,  
                                                        0.16645033,  0.12447953,  0.07260076,  
                                                        0.03062996,  0.01459854 ]),
                                    args=boxsize),

                                    )
                                
                                
        for key, value in regression_tests.items():
            r,w = value["function"](data,value["args"],return_weights=True)
            if not numpy.all(self.test.withinAbsoluteValue(w-value["target"],0.0,1e-8)):
                raise ValueError("Regression test of smoothing.{0} failed ({1})".format(key, value["function"]))
        
        import scipy
        for key, value in scipy_regression_tests.items():
            r,w = value["function"](data,value["args"],return_weights=True, window_name=value["window_name"], options=3)
            if not numpy.all(self.test.withinAbsoluteValue(w-value["target"],0.0,1e-8)):
                raise ValueError("Regression test of smoothing.{0} failed ({1})".format(key, value["function"]))
        
    


import numpy as np


def calculate_reverse_indices(incount, full_index, sparse=False):
    """Calculates the indices that go into each bin in a histogram
        and returns an array of values where
        ri[index] is list of indices of bins in the source data that
        are part of the [index] bin in the histogram.
        Called from within grid_data only.
    """
    #Two arrays to consider here
    #The count in each bin, regardless of data validity (missing data, masks, etc.)
    #called count_full, and the full_index that places every element into the output array
    
    #We define two arrays, rii contains arrays of each for each output bin, where each
    #array is the list of indices that go into that bin.  rjj contains 
    
    array_length = incount.size
    shp = incount.shape
    #create the storage arrays
    rii = np.empty(array_length, dtype=np.ndarray)
    rjj = np.zeros(array_length, dtype=np.int)
    #flatten the incount array
    c = np.reshape(incount, incount.size)
    #calculate the unique values and their indices, and an array that can
    #be used to reconstruct the full_index
    (fi_value, fi_index, fi_inverse) = \
            np.unique(full_index, return_index=True, return_inverse=True)
    #count the elements that go into each bin
    bc_fi_inverse = np.bincount(fi_inverse)

    #IF we're in sparse mode, then we should index into a counter array, instead of the real fi_value array    
    if sparse:
        locations = np.arange(fi_value.size)
    else:
        locations= fi_value

#new inversion code
    temp = np.argsort(fi_inverse)
    counter=0
    for count,elem in enumerate(locations):
        rjj[elem] = bc_fi_inverse[count]
        rii[elem] = np.sort(temp[counter:counter+rjj[elem]])# np.where(fi_inverse==count)[0] 
        counter=counter+rjj[elem]
#The commented-out code should be the equivalent code, but it's slower than the where method above
#If there are problems, change the code back to calculate the full arrays as below.
#    #For each pair of value and its count, define an array
#    #of the appropriate (count) size
#    for loc, val in zip(locations, bc_fi_inverse):
#        if val > 0:
#            rii[loc] = np.zeros(val, dtype=int)
#    #no that we've allocated enough room
#    #loop through the inverse array, find its location in the main array outside
#    #using the fi_value lookup (we could have just used full_index, but not in sparse mode)
#
#
#    for count, val in enumerate(fi_inverse):
#        #I think this can be replaced by index=locations[val] after the definition above
#        if sparse is False:
#            index = fi_value[val]
#        else:
#            index = val
#        #rjj'th element of the riith element is set the location (count) of the value in fi_inverse, 
#        #which is equivalent to the location in the input array
#        rii[index][rjj[index]] = count
#        #a counter, to know which element we should write to next.
#        rjj[index] = rjj[index] + 1
    #finally, reshape, and output
    rii = np.reshape(rii, shp)
    rjj = np.reshape(rjj, shp)

    return rii, rjj


def calculate_statistics(full_index, weights,
                         bs, square=None, sparse=False, missing_data=None):
    """
    Calculates the mean, variance,and count of each bin indexed in full_index with data coming from 
    weights
    """
    
    #we have two data arrays here. The first is the 'full_index' that identifies
    #the bin into which each element goes, and the un-normalized weight of that element
    #in weights. In reality, 'weights' is the data we want to grid. The function we call
    #later (bincount) happens to consider them as weights to a different array
    #but in doing so calculates the statistics appropriately.
    
    #calculate the product of bin sizes to get the largest possible array size
    sz = 1
    for d in bs:
        sz = sz * d
    #now call numpy.unique asking for the 
    #fi_value = values that give a unique array
    #fi_index = the indices of the values in fi_value
    #fi_inverse = the location of each element in full_index in fi_value
    #can be used to reconstruct the full_index value OR used to bin properly later
    (fi_value, fi_index, fi_inverse) =\
            np.unique(full_index, return_index=True, return_inverse=True)
    
    #now deal with missing data if present. #If we find no real data (len(w[0])==0)
    #then our answer is pretty simple as the result is also all zeros (for variance and count),
    #or missing for mean.
    if missing_data:
        w = np.where(weights != missing_data)
        if len(w[0]):
            index = w[0]
        else:
            if sparse is False:
                m = np.zeros(bs)
                return (m + missing_data, m.copy(), m.copy())
            else:
                s = np.zeros(len(fi_value))
                return (s + missing_data, s.copy(), s.copy())
    #OR if the arrays are masked numpy arrays, then we perform the same analysis
    #but ask whether the mask is
    #just a single bool, then perform the same test as above
    #an array of bools, then invert it to get the data selection array
    #(mask==true -> bad data), (index=true -> good_data)
    elif isinstance(weights, np.ma.core.MaskedArray):
        if isinstance(weights.mask, np.bool_):
            if not weights.mask:
                index = slice(None, None, None)
            else:
                if sparse is False:
                    m = np.zeros(bs)
                    return (m + missing_data, m.copy(), m.copy())
                else:
                    s = np.zeros(len(fi_value))
                    return (s + missing_data, s.copy(), s.copy())
        elif isinstance(weights.mask, np.ndarray):
            index = ~weights.mask
    #OR there's no missing data, select all the data
    else:
        index = slice(None, None, None)
    #for convenient later, we perform a count on all of the data, regardless of 
    #it's validity
    count_full = np.bincount(fi_inverse[slice(None, None, None)])
    #No calculate the count (unweighted bincount) of the valid data
    #and the sum of the data in each bin (weighted by data)
    #and the sum of squares (weighted by data**2)
    try:
        count = np.bincount(fi_inverse[index])
        su = np.bincount(fi_inverse[index], weights=weights[index])
        var = np.bincount(fi_inverse[index], weights=weights[index] ** 2)
    #not sure what exception I'm expecting here, anymore. 
    except:
        count = 0. * count_full
        su = 0. * count_full
        var = 0. * count_full
    #if we DONT want sparse data then the array_length is product of all dimensions
    #and the shape is given in the binsize array bs
    #if we DO want sparse data, the array_length is the largest number in the unique inverse
    #array fi_inverse, which seems like it should eqyal the length of fi_value
    if sparse is False:
        array_length = sz
        shp = bs
    else:
        array_length = np.max(fi_inverse) + 1
        shp = np.max(fi_inverse) + 1
    #construct the output arrays
    m = np.zeros(array_length)
    c = np.zeros(array_length)
    cf = np.zeros(array_length)
    v = np.zeros(array_length)
    w = np.where(count != 0)[0]

    #for count_full we can't look for the number of valid elements, because there don't
    #have to be any, so we copy the data immediately
    if sparse is False:
        cf[fi_value] = count_full
    else:
        cf = count_full #no indexing -> straight copy 
    
    #if there is data (w.size>0) then save the data in the output, otherwise
    #the initialization to 0 is good enough.
    #If NOT sparse, then the indexing is stored in fi_value (where its count is 
    #nonzero
    #if YES sparse, then the indexing is just the locations of nonzero elements
    #which I think is every element in w.
    forward_index = None
    forward_index = fi_value
    if w.size:
        if sparse is False:
            index = fi_value[w]
        else:
            index = w
        #divide the sums by the counts to get means, 
        m[index] = su[w] / count[w]
        v[index] = var[w] / count[w]
        c[index] = count[w]
#        cf[index] = count_full[w]
    #if square is set, return the mean of the square, not the variance.
    if square is None:
        v = v - m * m
    else:
        pass
    #in NON sparse mode, we need to reshape the output to N dimensional arrays
    #otherwise shp is just the array length and nothing happens here.
    m = np.reshape(m, shp)
    v = np.reshape(v, shp)
    c = np.reshape(c, shp)
    cf = np.reshape(cf, shp)
    
    return (m, v, c, cf, forward_index)


def grid_data(data, bins, mn=None, mx=None,
                reverse_indices=None, square=None,
                sparse=False, missing_data=None):
    """Grids data into regular grid of bins
    data is an N*M numpy array, where:
    N is the number of data points,
    A<M is the number of dimensions we are binning over
    and D=M-A is the number of dimensions of data we bin.
    Bins can be a list of arrays of bin edges, or a list of
    bin sizes (integers). If bin edges are given, they are
    assumed to be monotonic and used as in bincount. If they
    are bin sizes, optional maximum (mx) and minimum (mn)
    values for all dimensions can be given. In this case,
    the edge bins (above mx, or below mn) contain the overflow data.
    By default, the function returns the mean value, the variance,
    and the counts of each bin. If reverse_indices is set to True,
    the indices in the data array (along the N dimension) of each
    element that contributes to every bin is stored in a variable
    length array in that bin.
    
    :param data: 2-dimensional(N,M) array of data to grid. 
            The first A columns are used as axes and require corresponding entries 
            in the bins list. The remaining D columns are data to be binned
    :type data: numpy array

    :param bins: list of bin edges (upper edge) or bin lengths for each axis in A
    :type bins: list
    
    :param mn: list of bin minima if the bins list contains integer number of bins
    :type mn: list

    :param mx: list of bin maxima if the bins list contains integer number of bins
    :type mx: list

    :param reverse_indices: logical to determine if the indices that are used in each grid box are calculated
    :type reverse_indices: bool

    :param square: logical to determine if the mean of the square data mean(X**2) is returned instead of variance (mean(X**2)-mean(X)**2)
    :type square: bool

    :param sparse: logical to determine if the sparse method is used instead and a 1D gridded dataset is created
    :type sparse: bool
    
    :param missing_data: list of missing data entries for each data array in D(better to mask arrays instead)
    :type missing_data: list


    :return: Mean, calculated mean value of each grid box
    :return: Variance, calculated variance of each grid box (or mean of square)
    :return: Count, number of elements in each grid box
    
    With reverse_indices=true
    
    :return: ri, array containing a list for each gridpoint containing the index back into D of each datapoint
    :return: rj, the length of each list in ri
    
    Types
    
    :rtype: tuple of (numpy (D+1)d, numpy (D+1)d, numpy (D+1)d) (no reverse_indices, no sparse)
    :rtype: tuple of (numpy 1d, numpy 1d, numpy 1d) (no reverse_indices, sparse)
    :rtype: tuple of (numpy (D+1)d, numpy (D+1)d, numpy (D+1)d,numpy (D+1)d,numpy (D+1)d) (reverse_indices, no sparse)
    :rtype: tuple of (numpy 1d, numpy 1d, numpy 1d,numpy 1d,numpy 1d) (reverse_indices, sparse)

    """
    #glossary
    #axis/axes = The first M rows of data that are used to filter the data into a grid
    #bins = The values of the axes that are used to group elements
    #data = The non-filterable part of the input data that is grouped by axes into bins
    
    import numpy as np
    #get the shape of the incoming data, create a bin size array read for bin data
    s = data.shape
    bs = np.zeros(len(bins), dtype=np.int)
    #if the first element is a numpy array, they should all be
    #then I assume that each element in bins contains the upper bin edges
    #of each dimension.
    if type(bins[0]) == np.ndarray:
        #if arrays are given in 'bins', assume they are bin edges
        use_bins = bins
        for i in range(len(bs)):
            bs[i] = len(use_bins[i])
            pass
        pass
    else:
        #if non-lists (int,long,float) are given, assume they are nbins
        #convert them to longs
        bs = [int(b) for b in bins]

        #if min or max are not given, calculate them
        if mn is None:
            mn = np.min(data, 0)
        if mx is None:
            mx = np.max(data, 0)
        use_bins = []
        for i in range(len(bs)):
            use_bins.append(
                np.linspace(mn[i], mx[i], bs[i] + 1)
                )
        pass

    #now we have bin edges for each of the axes, either supplied or generated
    #Calculate the product of the bin sizes
    sz = 1
    for d in bs:
        sz = sz * d
    #Create the index array that is the same shape as the incoming data
    #except for one bin in the last dimension,
    #TODO: I think the s[1]-1 dimension could be len(bs)
    #index = np.zeros((s[0], s[1] - 1), dtype=np.int)
    #implementing the TODO above
    index = np.zeros((s[0], len(bs)), dtype=np.int)
    #for each index array, call numpy.digitize on the data which calculates
    #the index into the bin array that each element of the axis data appears in.
    #We only use n-1 elements of each bin so that the overflow appears in the last
    #bin, not in the (undefined) bin above the last defined one.
    for i in range(len(bs)):
        index[:, i] = np.clip(bs[i] - np.digitize(data[:, i],
                            use_bins[i][::-1]),
                            0, bs[i] - 1)

    #Now we have the indexes into each dimension, combine them into a single index
    #by scaling the full_index by the product of the prior dimensions and adding the
    #next index. This is the same way you construct a multi-dimension array out of
    #one dimensional data.
    full_index = index[:, 0]
    for i in range(1, len(bs)):
        full_index = full_index * bs[i] + index[:, i]

    lb = len(bins)
    r = data.shape[1] - lb
    #if there is only one 'data' row (i.e. 1 non-filterable row) then we need to
    #call the calculate_statistics function with the correct shape
    #OR we loop through the data and call the function for each row.
    if r==0:
        (m, v, c, cf,forward_index) = calculate_statistics(full_index, data[:, -1]*0., bs,
                                        square=square, sparse=sparse
                                        )
        count = c
        count_full = cf
    elif r == 1:
        (m, v, c, cf,forward_index) = calculate_statistics(full_index, data[:, -1], bs,
                                        square=square, sparse=sparse,
                                        missing_data=missing_data)
        count = c
        count_full = cf
    else:
        #if there's no missing data, generates a list of 'None's to fool the algorithm,
        #otherwise there must be the correct number of missing data values for the
        #size of the data. Also acceptable is to pass in a masked array instead,
        #with no missing data.
        if missing_data is None:
            missing_data = [None for i in range(lb, data.shape[1])]
        for i in range(lb, data.shape[1]):
            (m, v, c, cf, forward_index) = calculate_statistics(full_index, data[:, i], bs,
                                            square=square, sparse=sparse,
                                            missing_data=missing_data[i - lb])
            #If we wanted non-sparse data, m will be Naxis dimensional, so generate
            #a slice-index thing for that shape
            #otherwise generate a single dimension slice, with a blank axis at the
            #end to allow concatenation
            slall = slice(None, None, None)
            if sparse is False:
                sl = [slall for j in range(len(m.shape))]
                sl.append(np.newaxis)
            else:
                sl = [slall, np.newaxis]
                pass
            #For the first element i==lb we need to create the final storage variables
            #for other elements i>lb , we need to concatenate
            if i == lb:
                m2 = m[sl]
                v2 = v[sl]
                c2 = c[sl]
                count = c
                count_full = cf
            else:
                m2 = np.concatenate((m2, m[sl]), axis=-1)
                v2 = np.concatenate((v2, v[sl]), axis=-1)
                c2 = np.concatenate((c2, c[sl]), axis=-1)
                pass
            pass

        m = m2
        v = v2
        c = c2
    #if we don't want to calculate the reverse indices array, we can stop here
    #otherwise we need to call the appropriate function and add the results
    #to the tuple we return
    result = [m,v,c]
    if sparse:
        result.append(forward_index)
    if reverse_indices:
        ri, rj = calculate_reverse_indices(count_full, full_index, sparse=sparse)    
        result.extend([ri,rj])
    return result



"""Interal fft code. Interfaces to anfft, numpy.fft or scipy.fft for use by fft"""

import numpy
got_fft_library=False
using_fft_library=None
#Try to get an FFT library in order of anfft, scipy.fftpack, numpy.fft
#we require numpy for everything else, so the last check shouldn't fail?

if not got_fft_library:
    try:
        import anfft as fft
        got_fft_library = True
        using_fft_library="anfft"
    except ImportError:
        got_fft_library = False

if not got_fft_library:
    try:
        import scipy
        import scipy.fftpack as fft
        got_fft_library = True
        using_fft_library="scipy"
    except ImportError:
        got_fft_library = False

if not got_fft_library:
    try:
        import numpy.fft as fft
        got_fft_library = True
        using_fft_library="numpy"
    except ImportError:
        got_fft_library = False

if not got_fft_library:
    raise ImportError("Error: Cannot use fft libraries")        

import numpy.fft as npfft

fftshift = npfft.fftshift
def fft_ifftn(*args, **kwargs):
    """Call the core fft library function ifftn (inverse)
    
    If the anfft library is used, enable 'measure' optimization.
    
    """
    if using_fft_library=="anfft":
        return fft.ifftn(*args, measure=True, **kwargs)
    else:
        return fft.ifftn(*args, **kwargs)

def fft_fftn(*args, **kwargs):
    """Call the core fft library function fftn.
    
    If the anfft library is used, enable 'measure' optimization.
    
    """
    
    if using_fft_library=="anfft":
        return fft.fftn(*args, measure=True, **kwargs)
    else:
        return fft.fftn(*args, **kwargs)
    
def odd(n):
    """Returns true if the input is odd
    
    :param n: number to test
    :type n: integer
    
    :return: logical if input is odd or not
    :rtype: Boolean
    
    """
    if n % 2 == 1:
        return True
    else:
        return False


def rffttruncate(ft, axis):
    """Truncates the FFTd data along the particular axis
    
    :param ft: multiple dimension array with at least one dimension FFT'd
    :type ft: numpy array

    :param axis: axis to truncate
    :type axis: list

    :return: Truncated data
    :rtype: numpy array

    This function shifts (rolls) the data along the correct axis, calculates a slice range
    which is either (N-1)/2 or N/2-1.
    
    """
    ft2 = numpy.roll(ft, -1,axis)
    d=[]
    for i in range(ft.ndim):
        d.append(slice(None,None,None))

    n=ft2.shape[axis]
    if odd(n):
        x = (n-1)/2
    else:
        x=n/2-1
    d[axis]=slice(-1,x-1,-1)
    return ft2[d]

def ffttruncate(ft, axis):
    """Truncates the FFTd data along the particular axis
    
    :param ft: multiple dimension array with at least one dimension FFT'd
    :type ft: numpy array
    
    :param axis: axis to truncate
    :type axis: list
    
    :return: Truncated data
    :rtype: numpy array
    
    This function shifts (rolls) the data along the correct axis, calculates a slice range
    which is either (N-1)/2 or N/2-1.
    
    """
    d=[]
    for i in range(ft.ndim):
        d.append(slice(None,None,None))

    n=ft.shape[axis]
    x = (n+1)/2
    d[axis]=slice(None,x,None)
    return ft[d]

def fftshift(ft, axis):
    """Calls the fftshift core function, which shifts the ft half way along the axis
    :param ft: array to be shifted
    :type ft: numpy array
    
    :param axis: axis to shift over
    :type axis: integer
    
    :return: shifted array
    :rtype: numpy array
    
    This function is only available in numpy
    
    """
    return npfft.fftshift(ft,axis)


def fft_axes(ft, d1, d2, axis):
    """Uses numpy to generate the axes because on step size and length
    
    :param ft: Input fft'd data, used only for the dimension data
    :type ft: numpy array
    :param d1: dimension 1 step size
    :type d1: float
    :param d1: dimension 2 step size
    :type d2: float
    :param axis: axis indexes of dimensions
    :type axis: list
    
    :return: tuple of axes values
    :rtype: tuple of integers
    """
    n=ft.shape[axis[0]]
    f1 = npfft.fftfreq(n, d1)
    n=ft.shape[axis[1]]
    f2 = npfft.fftfreq(n, d2)

    return (f1, f2)


def trans_array(data, _axes):
    """Calculates indices for an array transpose use by anFFT.
    
    :param data: The data, used for dimension purposes (rank)
    :type data: numpy array
    :param _axes: axes to FFT over, and hence transpose to the end
    :type _axes: list
    :return: tuple of axes values
    :rtype: tuple of integers
    
    """
    ind = list(range(data.ndim))
    try:
        axes = list(_axes)
    except: #integer
        axes=[_axes]
    axes.sort()
    axes.reverse()
    for a in axes:
        ind.pop(a)
    ind.extend(axes)
    return ind

def rtrans_array(data, _axes):
    """Calculates the indices to reverse the transpose of an array after anFFT.
    
    :param data: The data, used for dimension purposes (rank)
    :type data: numpy array
    
    :param _axes: axes to FFT over, and hence transpose to the end
    :type _axes: list
    
    :return: tuple of axes values
    :rtype: tuple of integers
    """
    ind = trans_array(data, _axes)
    rind=list(ind)
    for i,j in enumerate(ind):
        rind[j]=i
    return rind

def trans_forward(data, _axes):
    """Transpose an array to the correct order for anfft
    
    :param data: The data to be transposed
    :type data: numpy array
    
    :param _axes: axes to FFT, and hence transpose
    :type _axes: list
    
    :return: The transposed array
    :rtype: numpy array
    """
    ind = trans_array(data, _axes)
    #ind now contains the transposed dimensions
    return numpy.transpose(data, ind)
    
def trans_backward(data, _axes):
    """Reverse the transpose of an array to the correct order for anfft
    
    :param data: The data to transpose
    :type data: numpy array
    
    :param _axes: The axes over which to FFT, and transpose if anfft is used
    :type _axes: list
    
    :param function: ifft function, or fft function if overridden
    :type function: function pointer
    
    :return: The transposed array
    :rtype: numpy array
       """
    ind = trans_array(data, _axes)
    ind2=list(ind)
    for d in ind:
        ind2[d]=ind.index(d)
    #ind now contains the reverse transposed dimensions
    return numpy.transpose(data, ind2)

    
def ifftn(data,axes=None, function=None):
    """Runs the inverse FFT function over multiple axes
    
    :param data: The data to transpose
    :type data: numpy array
    
    :param axes: The axes over which to FFT, and transpose if anfft is used
    :type axes: numpy array

    :param function: ifft function, or fft function if overridden
    :type function: function pointer

    :return: An FFTd array 
    :rtype: numpy array
    
    """
    if function is None:
        function =fft_ifftn
        
    if axes is None:
        return function(data)
    else:
        if using_fft_library == "anfft":
            #transpose the array so that the fft dims are all at the end
            data = trans_forward(data, axes)
            #fft
            if type(axes) is list:
                k = len(axes)
            elif type(axes) is numpy.ndarray:
                k = axes.size
            elif type(axes) is int:
                k=1
            ft = function(data, k)
            #transpose back
            data = trans_backward(data, axes)
            ft = trans_backward(ft, axes)
            
            pass
        elif using_fft_library == "scipy":
            ft = function(data, axes=axes)
            pass
        elif using_fft_library == "numpy":
            ft = function(data, axes=axes)
            pass
        else:
            print("Error: Can't get here")

        return ft

def fftn(data,axes=None):
    """Runs the inverse FFT function over multiple axes
    
    :param data: The data to transpose
    :type data: numpy array

    :param axes: The axes over which to FFT, and transpose if anfft is used
    :type axes: numpy array

    :param function: ifft function, or fft function if overridden
    :type function: function
    
    :return: An FFTd array 
    :rtype: numpy array
    """
    return ifftn(data, axes, fft_fftn)

    




#http://stackoverflow.com/questions/2082152/case-insensitive-dictionary
class CaseInsensitiveDict(dict):
    @classmethod
    def _k(cls, key):
        return key.lower() if isinstance(key, str) else key
    def __init__(self, *args, **kwargs):
        super(CaseInsensitiveDict, self).__init__(*args, **kwargs)
        self._convert_keys()
    def __getitem__(self, key):
        return super(CaseInsensitiveDict, self).__getitem__(self.__class__._k(key))
    def __setitem__(self, key, value):
        super(CaseInsensitiveDict, self).__setitem__(self.__class__._k(key), value)
    def __delitem__(self, key):
        return super(CaseInsensitiveDict, self).__delitem__(self.__class__._k(key))
    def __contains__(self, key):
        return super(CaseInsensitiveDict, self).__contains__(self.__class__._k(key))
    def has_key(self, key):
        return key in self
    def pop(self, key, *args, **kwargs):
        return super(CaseInsensitiveDict, self).pop(self.__class__._k(key), *args, **kwargs)
    def get(self, key, *args, **kwargs):
        return super(CaseInsensitiveDict, self).get(self.__class__._k(key), *args, **kwargs)
    def setdefault(self, key, *args, **kwargs):
        return super(CaseInsensitiveDict, self).setdefault(self.__class__._k(key), *args, **kwargs)
    def update(self, E={}, **F):
        super(CaseInsensitiveDict, self).update(self.__class__(E))
        super(CaseInsensitiveDict, self).update(self.__class__(**F))
    def _convert_keys(self):
        for k in list(self.keys()):
            v = super(CaseInsensitiveDict, self).pop(k)
            self.__setitem__(k, v)

CaselessDict=CaseInsensitiveDict


#
def data_filename(modname, filename):
    """Given the module name, and filename, finds the path to the file from the python sys.modules
    dictionary. Used to access the header dictionary file"""
    import os, sys
    filename = os.path.join(
        os.path.dirname(sys.modules[modname].__file__),
        filename)
    return filename
    



import numpy as np
import netCDF4
from .boolfilter import boolfilter
import logging
import dwell.io.netcdf as netcdf

def filter_time(time, start_date, end_date):
    if start_date is None:
        start_date = -9e36
    if end_date is None:
        end_date = 9e36
    return (time >=start_date)&(time<end_date)
    
def find_indices_to_extract(ncin, time_axis, start_date=None, end_date=None, 
                            filter_condition=None, filenames=None, log=None,
                            keep_duplicates=False):
    """This function finds the indices to extract from each NetCDF filehandle in ncin 
    using the start_date and end_date."""
    if log is None:
        log = logging.getLogger("pydwell")
    
    indices=[]
    times=[]
    count=0
    fcount=0
    for nc in ncin:
        log.debug("Filtering file {0} of {1}".format(fcount,len(ncin)))
        time = nc.variables[time_axis][:]
        if filter_condition is None:
            w=np.where(filter_time(time,start_date, end_date))
        else:
            w=np.where(boolfilter(nc, filter_condition))
        
        if len(w[0]):
            indices.append(w[0]) #return w[0]
            times.append(time[w[0]])
            count=count+len(w[0])
        else:
            indices.append([]) #return []
        fcount = fcount + 1
    #print(count, time_axis)
    if count > 0:
        indices, ncin, count, redundant_ncin = filter_duplicate_times(ncin, time_axis, indices, count, filenames=None, keep_duplicates=keep_duplicates)
    return indices, ncin, count, redundant_ncin

def filter_duplicate_times(ncin, time_axis, indices, count, filenames = None, time_threshold=1.0, log=None, keep_duplicates=False):
    """
        Filter the times from ncin to find duplicates.
        1. Starts by finding the minimum and maximum times in each of the available files.
        2. Sorts these files based on the start_time.
        3. For each of the available files, collect as many indices from the file
        using the early file in preference to later files such that overlapping data 
        from early files is taken instead of later files, and overlapped data in 
        later files is ignored.
    """
    if filenames is None:
        filenames = [""]*len(ncin)
    if log is None:
        log = logging.getLogger("pydwell")
    #first of all, find the min and max valid times from each files
    time_range=[]
    for nc, ind in zip(ncin, indices):
        if len(ind):
            time = nc.variables[time_axis][ind]
            time_range.append( [time[0], time[-1]] )
        else:
            time_range.append( [] )
    #The time_range array either has an empty list for each file, or the start and end times of each file.
    #now we sort files based on min_time -> time_range[][0]
    #sort function is the first element of the first element of the tuple.
    #i.e. we provide a tuple of elements to sort, the first is the time range tuple.
    sort_function = lambda tup: tup[0][0]
    new_time, new_nc, new_ind, new_filenames = list(zip(*sorted(zip(time_range,ncin, indices,filenames))))
    result = []
    #Now we're sorted, iterate through the files, creating the results list which contains
    #a list corresponding to the files
    #where each entry is a list of indices, netcdf filehandle, filename, and times in epoch format
    #the last of these entries is the time, where the first element is the first element taken from the file
    #the last element is the last element from the file (confusing no..)
    redundant_ncin=[]
    for t, nc, ind,f in zip(new_time, new_nc, new_ind,new_filenames):
        if len(t) ==0: #no new data, don't do anything
            redundant_ncin.append(nc)
            continue
        #at the start we are empty
        if len(result): #have some data already in the list
            #This element (the last of the last of the last) is
            #the last element of the time element of the previous file
            last_time = result[-1][-1][-1] 
            new_file_time = nc.variables[time_axis][ind]
            #keep data that is later than this observation by at least time_threshold seconds
            filter_func = lambda tup: tup[0] > (last_time+time_threshold/86400.)
            if keep_duplicates:
                filter_func = lambda x: True
            
            new_data = list(filter(filter_func, list(zip(new_file_time, ind))))
            if len(new_data):
                new_time, new_ind = list(zip(*new_data))
                result.append([np.array(new_ind),nc,f,t])
            else: #no data to append
                redundant_ncin.append(nc)
                pass
        else:
            result.append([ind, nc, f, t])
            
    #finish the loop
    count = sum([len(x[0]) for x in result])
    newind, newnc, newfilenames, newtime = list(zip(*result))
    return newind, newnc, count, redundant_ncin
    
def select_data(var, newtime):
    """select the new data from var corresponding to indices given in newtime"""
    #if index of time is 0 only
    #thats annnnoying
    if len(var.shape) >= 2:
        return var[newtime]
    elif len(var.shape) ==1: 
        #print var[newtime]
        return var[newtime]
    else:
        raise Exception("Bugger, I have no clue why this happened")

def extract_field(ncin, newtime, varname, aslist=False, log=None):
    """Extract the var with name=varname from the ncin NetCDF files, using the corresponding
    time axis stored in newtime"""
    if log is None:
        log = logging.getLogger("pydwell")
    data=[]
    #do we have the appropriate dimensions to filter this code
#    if ncin[0].variables[vari
    for nc, nt in zip(ncin, newtime):
        if len(nt) > 0:
            data.append(select_data(nc.variables[varname], nt))
    if not aslist:
        data = np.concatenate(data,0)#-1)
    return data


def create_dimensions(ncout, time_axis, canonical_ncin, log=None):
    """Create all of the necessary dimensions in ncout using the canonical_ncin as a template"""
    if log is None:
        log = logging.getLogger("pydwell")
#    log.debug("Creating Dimensions in NetCDF")
    dim={}
    obs_variables = [x for x in list(canonical_ncin.dimensions.keys()) if not x.startswith(time_axis)]
                
    
    for dname in obs_variables:
        dval = canonical_ncin.dimensions[dname]
#        log.debug("Obs Dimension {0} has size {1}".format(dname, len(dval)))
        dim[dname] = ncout.createDimension(dname, len(dval))
    
    return dim

def create_variables(ncout, time_axis, canonical_ncin, ncin, newtime, log=None, zlib=False, filter_variables=None):
    """Create the necessary variables from the canonical_ncin NetCDF file, then extract
    the data from each of the ncin files as needed and save the data in the output variable.
    """
    if log is None:
        log = logging.getLogger("pydwell")
#    pbar = progressbar.ProgressBar(len(canonical_ncin.variables.keys())).start()
    counter=0
    variables = canonical_ncin.variables
    #filter out data with prefixes
    #filter out data with no time index as the first index.
    #filter(lambda x: canonical_ncin.variables[x].dimensions[0] == time_axis, variables)
    use_variables = [x for x in variables if canonical_ncin.variables[x].dimensions[0] == time_axis]
    #filter(lambda x: time_axis not in canonical_ncin.variables[x].dimensions, variables)
    copy_variables = [x for x in variables if time_axis not in canonical_ncin.variables[x].dimensions]
    netcdf.copy_variables(canonical_ncin, ncout, include=copy_variables)
    if filter_variables is not None:
        #use_variables = filter(lambda x: x in filter_variables,use_variables)
        use_variables = [x for x in use_variables if x in filter_variables]
    
    for varname in use_variables:
        log.debug("Variable {0}".format(varname))
        canonical_var = canonical_ncin.variables[varname]
#        pbar.update(counter)
        counter=counter+1
        #collect data for this variable
        data=extract_field(ncin, newtime, varname)
        mydim = canonical_var.dimensions
        #att = dict(
        #            map(
        #                lambda x: (x, getattr(canonical_var,x)),canonical_var.ncattrs()
        #                )
        #        )
        att = [(x,getattr(canonical_var,x)) for x in canonical_var.ncattrs()]
        mytype = canonical_var.dtype
        if "_FillValue" in canonical_var.ncattrs():
            newvar = ncout.createVariable(varname, mytype, mydim, fill_value = getattr(canonical_var, "_FillValue"), zlib=zlib)#
        else:
            newvar = ncout.createVariable(varname, mytype, mydim, zlib=zlib)#
        newvar[:] = data[:] # attributes=att)
        for attname, attval in six.iteritems(att):
            if attname != "_FillValue":
                setattr(newvar,attname, attval)

  
def filter_netcdf(input_files, output_file, time_axis="Time", low_time=None, high_time=None, filter_condition=None, zlib=False, filter_variables=None,log=None,keep_duplicates=False):
    """Filter the NetCDF input files based on the time_axis entry from 'low' to 'high' and create
    a new file out of the input files and restrictions. Default values for 0<=time<=9e36 ."""
    infiles = [netCDF4.Dataset(input,'r') for input in input_files]
    output = netCDF4.Dataset(output_file,'w')

    if filter_condition is None and low_time is None:
        low_time = -9e36
    if filter_condition is None and high_time is None:
        high_time = 9e36
    
    indices, ncin, count, redundant = find_indices_to_extract(infiles, time_axis, start_date=low_time, end_date=high_time, 
                                                              filter_condition=filter_condition, filenames=input_files,log=log, keep_duplicates=keep_duplicates)
    
    output.createDimension(time_axis, count)
    #create dimensions from first input
    create_dimensions(output, time_axis, ncin[0],log=log)
    create_variables(output, time_axis, ncin[0], ncin, indices, zlib=zlib, filter_variables=filter_variables,log=log)
    for x in redundant:
        x.close()
    for x in ncin:
        x.close()
    output.close()
        



"""Terminal prompt utilities"""
import six
# prompt
def prompt(text=None, choices=None, default=None):
    """Create a prompt of text is given,then offer choices with a default.
    e.g. prompt('how are you?',['good','bad'],['good']"""

    if text is None:
        text = ""

    use_choices = None
    if choices is not None:
        if type(choices) is list:
            use_choices = dict(zip(choices, choices))
        elif type(choices) is dict:
            use_choices = choices

    while True:
        if use_choices is None:
            ans = six.input("{text}: ".format(text=text))
            if not ans:
                return default
            else:
                return ans

        else:  # choices
            ans = six.input("{text} [{choices}]: ".format(text=text,
                                                    choices=",".join(
                                                    use_choices.keys())))
            for k in use_choices.keys():
                if (type(use_choices[k]) == str and ans == use_choices[k]) or\
                (type(use_choices[k]) == list and ans in use_choices[k]):
                    return k
            #no choice found
            if default is not None:
                return default

            message("Please enter one of the choices.")
            continue


def yesno(text=None):
    """Wrapper around prompt for a simple yes/no question"""
    choices = {"y": ["YES", "yes", "y", "Y"],
               "n": ["NO", "no", "n", "N"]}

    res = {"y": True, "n": False}
    return res[prompt(text, choices=choices, default=None)]



"""FFT interface fo' science"""

try:
    import cdms2
    use_cdms2 = True
except:
    use_cdms2 = False
    
from ._internal import *

import numpy

def spec1d(data, d1, use_axes=0):
    """Calculate the 1 dimensional FFT and return a truncated array
    
    :param data: 1-dimensional array of data to FFT
    :type data: numpy array

    :param d1: : step size of the data, used to calculate axes data
    :type d1: float
    
    :return: Amplitude, calculated as the absolute magnitude of the complex FFT
    :return: phase, calculate as atan(imag(ft)/real(ft))
    :return: x axis values
    :rtype: tuple of (numpy 1d, numpy 1d, numpy 1d)
    """
    internal_use_axes =list([use_axes])
    ft = ifftn(data, internal_use_axes)
    n=ft.shape[internal_use_axes[0]]
    f1 = npfft.fftfreq(n, d1)
    f1 = abs(rffttruncate(f1,0))
    ft = rffttruncate(ft,internal_use_axes[0])
    #double amplitude and correct 0,0
    ft=ft*2.0
    ft[0]=ft[0]*0.5

    amp = numpy.abs(ft)
    pha = numpy.arctan2(numpy.imag(ft), numpy.real(ft))
    return (amp, pha, f1)
    
    
def axis_or_axes(data, axis=None, axes=None,*args, **kwargs):
    """Return either axes, axis or the indices of the last two dimensions
    
    :param axis: indices to use in the fft
    :type axis: list
    
    :param axes: alternative naming
    :type axes: list
    
    :return: axis is returned if defined, otherwise axes is defined
             otherwise the data rank is found, and the last two 
             dimensions are returned as indices
    :rtype: list
    """
    r = list(axis or axes or [data.ndim-2, data.ndim-1])
    return r

  

def truncate_fft(ft, d1, d2, use_axes, atmosphere=None, ocean=None):
    """Truncates a Complex 2d FFT plane to half the size in one dimension, removing negative exponents
    
    Given a 2-D NxM complex fft data, and step sizes along each, calculate the axis values 
    based on the step sizes, then truncate the 2D plane and axes to the right shape based on flags
    
    :param ft: 2D FFT'd data
    :type ft: numpy array
    
    :param d1: dimension along the X axis
    :type d1: float
    
    :param d2: dimension along the Y axis
    :type d2: float
    
    :param use_axis: two element axes list describing which axes are being used in the FFT, other axes are simply looped over
    :type use_axis: logical
    
    :param atmosphere: bool to decide to use 'atmosphere' truncation (f1 positive definite)
    :type atmosphere: logical

    :param ocean: bool to decide to use 'ocean' truncation (f2 positive definite)
    :type ocean: logical
    
    :return: tuple of complex truncated fft, axes1, axes2
    :rtype: 2d numpy, 1d numpy, 1d numpy
    """
    ao = False
    if atmosphere is not None:
        ao=True
    elif ocean is not None:
        ao=not ocean

    (f1, f2) = fft_axes(ft, d1, d2, use_axes)
    if ao: 
        #atmosphere
        #for atmosphere, wave (d1,f1) is positive definite
        #which means freq (d2,f2) can be negative

        f1 = -rffttruncate(f1, 0)
        ft = rffttruncate(ft,use_axes[0])
        #double amplitude and correct 0,0
        ft=ft*2.0
        ft[0,0]=ft[0,0]*0.5
        f2 = fftshift(f2, 0)
        ft = fftshift(ft,use_axes[1])
        
        pass
    else:
        #ocean

        f2 = -rffttruncate(f2,0)
        ft = rffttruncate(ft,use_axes[1])
        ft=ft*2.0
        ft[0,0]=ft[0,0]*0.5
        f1 = fftshift(f1, 0)
        ft = fftshift(ft, use_axes[0])
        pass

    return (ft, f1, f2)


def fft_core(field1, field2=None, use_axes=None):
    """Call the FFT core interface given data and axes to transform over
    
    :param field1: dataset 1, must contain at least the axes listed in use_axes
    :type field1: 2d numpy
    
    :param field2: optional extra field to be transformed. If field2 is not None, then the cospectrum of
                   field1 and field2 are calculated, as fft(field1)*conj(fft(field2))
    :type field2: 2d numpy
    
    :param use_axes: axes over which to FFT.
    :type use_axes: list

    :return: The fft'd field, or the cospectrum field
    :rtype: 2d numpy array
    """
    ft = ifftn(field1, use_axes)
#cospectrum optional
    if field2 is not None:
        ft2 = ifftn(field2, use_axes)
        ft = ft*numpy.conj(ft2)

    return ft

def spec(_data, d1, d2, *args, **kwargs):
    """Interface to spec_cdms or spec_numpy to wrap the result appropriately"""
    if use_cdms2:
        if isinstance(_data, cdms2.tvariable.TransientVariable):
            return spec_cdms2(_data,d1,d2,*args, **kwargs)
        elif isinstance(_data, numpy.ndarray):
            return spec_numpy(_data,d1,d2,*args, **kwargs)
    else:
        if isinstance(_data, numpy.ndarray):
            return spec_numpy(_data,d1,d2,*args, **kwargs)
    
def spec_cdms2(_data, d1, d2, *args, **kwargs):
    """Calls the fft spec_numpy function, then wraps the output
    in cdms variables to return a pretty object to the user."""
    
#cdms variable
    #break the variable into data and axes
    (amp,pha,freq,wave) = spec_numpy(_data.data,d1,d2,*args,**kwargs)

    #reconstruct two transient variables 
    _data=0.0
    amp_var = cdms2.tvariable.TransientVariable(amp)
    amp=0.0
    pha_var = cdms2.tvariable.TransientVariable(pha)
    pha=0.0
    wave_var = cdms2.tvariable.TransientVariable(wave)
    freq_var = cdms2.tvariable.TransientVariable(freq)
    #find out which axes are being used
    use_axis = axis_or_axes(_data,*args, **kwargs)
    not_fft_axes = set(numpy.arange(_data.ndim)).difference(set(use_axis))

    for a in not_fft_axes:
        amp_var.setAxis(a,_data.getAxis(a))
        pha_var.setAxis(a,_data.getAxis(a))

    #add axes
    wave_axis = cdms2.axis.TransientAxis(wave,id="wavenumber")
    freq_axis = cdms2.axis.TransientAxis(freq,id="frequency")
    for v in [amp_var,pha_var]:
        v.setAxis(use_axis[0], freq_axis)
        v.setAxis(use_axis[1], wave_axis)
    
    freq_var.setAxis(0, freq_axis)
    wave_var.setAxis(0, wave_axis)
    
    return (amp_var, pha_var, wave_var, freq_var)


def spec_numpy(_data, d1, d2,
              axis=None, axes=None,
              detrend=None, perturb=None,
              atmosphere=None, ocean=None, field2=None
              ):
    """Routine transforms a real field to frequency/zonal wavenumber space
    Positive k are eastward propagating
    Negative k are westward propagating
    frequency is always positive
    
    Phase has been defined as im(wave)/re(wave)
    wave = exp(i(k.lon - w.t + phi))
    and c_phase = w/k, w=s.Omega, s = freq = 1./(period/days)
    t + lon/omega is the local time at longitude lon
    for this convection local time of first max. = phi/w
    longitude of first max.  = -phi/k
    
    :param _data: The data to transform
    :type _data: 2d numpy
    
    :param d1: axis 1 step size
    :type d1: float
    
    :param d2: axis 2 step size
    :type d2: float
    
    :param axes: axis over which to fft
    :type axes: list
    
    :param axis: alternate name for axes
    :type axis: list
    
    :param detrend: boolean to determine whether the data is first detrended over some axis NOT USED
    :type detrend: logical
    
    :param perturb: boolean to determine whether the data is first detrended over the other axis NOT USED
    :type perturb: logical
    
    :param atmosphere: booleans to determine the axis to truncate
    :type atmosphere: logical

    :param ocean: booleans to determine the axis to truncate
    :type ocean: logical
    
    :param field2: optional argument to flag a cospectrum calculation.
    :type field2: float
    """

    data = _data
    #which axes?
    use_axes = axis_or_axes(data,axis,axes)
    #calculate the fft over n dimensions
    ft = fft_core(data, field2=field2, use_axes=use_axes)
    #calculate axis and truncate    
    ft,f1,f2 = truncate_fft(ft, d1, d2, use_axes, atmosphere, ocean)

    if field2 is None:
        #amplitude spec
        amp = numpy.abs(ft)
        pha = numpy.arctan2(numpy.imag(ft),numpy.real(ft))
        return (amp,pha,f1, f2)
    else :
        #cospectrum
        cospec = numpy.real(ft)
        quad = numpy.imag(ft)
        return (cospec,quad,f1, f2)

def cospec(_data, field2, d1, d2,
            axis=None, axes=None,
            detrend=None, perturb=None,
            atmosphere=None, ocean=None
            ):
    """Calculates the cospectrum as fft(_data)*conj(fft(field2))
    
    :param _data: first field
    :type _data: 2D numpy array

    :param field2: second field
    :type field2: 2D numpy array
    
    :param d1: axis 1 step size
    :type d1: float
    :param d2: axis 2 step size
    :type d2: float
    
    :param axes: axis over which to fft
    :param axis: alternate name for axes
    :param detrend: boolean to determine whether the data is first detrended over some axis NOT USED
    :param perturb: boolean to determine whether the data is first detrended over the other axis NOT USED
    :param atmosphere,ocean: booleans to determine the axis to truncate.
    """
    return spec(_data, d1, d2,
    axis=axis, axes=axes,
    detrend=detrend, perturb=perturb,
    atmosphere=atmosphere, ocean=ocean, field2=field2
    )



from dwell.RAD.deprecated import deprecated

def new1():
    print("new1")

@deprecated
def dep1():
    print ("dep1")

@deprecated(message="new message.")
def dep2():
    print ("dep2")

@deprecated(message="new message.", replacement=new1)
def dep3():
    print ("dep2")

@deprecated(replacement=new1)
def dep4():
    print ("dep3")
#
@deprecated(replacement=new1, use_replacement=True)
def dep5():
    print ("dep4")

print("Running dep1")
dep1()
print("Running dep2")
dep2()
print("Running dep3")
dep3()
print("Running dep4")
dep4()
print("Running dep5")
dep5()
print("Running new1")
new1()



from setuptools import setup, find_packages
import sys, os

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
NEWS = ""#open(os.path.join(here, 'NEWS.txt')).read()


version = '1.0.0a1'

install_requires = [
    "numpy",
    "scipy",
    "termcolor",
    "six",
    "argh"
    # List your project dependencies here.
    # For more details, see:
    # http://packages.python.org/distribute/setuptools.html#declaring-dependencies
]


setup(name='dwell',
    version=version,
    description="Some Python packages for fft, smoothing, gridding",
    long_description=README + '\n\n' + NEWS,
    keywords='',
    author='Christopher Lee',
    author_email='lee@aeolisresearch.com',
    url='http://christopherlee.co.uk',
    license='BSD',
      classifiers=['Development Status :: 3 - Alpha',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   ],
    packages=find_packages('src'),
    package_dir = {'': 'src'},include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    entry_points={
        'console_scripts':
            ['dwell=dwell:main']
    }
)



import nose.tools
from dwell.caselessdict import CaselessDict
from dwell.testing import Testing

class test_CaselessDict(object):
    def __init__(self):
        self.test = Testing()
        return
    
    def test_null(self):
        self.test.debug("null")
        assert(True)
        
    def test_init(self):
        self.test.debug("empty init")
        a=CaselessDict()
        self.test.debug("copy init1")
        a=CaselessDict({"a":1,"b":2})
        self.test.debug("copy init2")
        a=CaselessDict({"a":1,"A":2})

    def test_getset(self):
        a=CaselessDict()
        self.test.debug("set")
        a["a"]=1
        self.test.debug("get")
        b=a["a"]
        assert(b==a["a"])
        self.test.debug("contains")
        assert("a" in a)
        assert(not "b" in a)
        
        self.test.debug("get2")
        assert(a.get("b",2)==2)
        assert(not a.get("b",3)==2)
        self.test.debug("setdefault")
        assert(a.setdefault("c",3)==3)
        
    def test_addondict(self):
        self.test.debug("update")
        a=CaselessDict()
        a.update({"a":1,"b":2})
        assert("a" in a and "b" in a)
        a.update(CaselessDict({"A":2,"c":3}))
        assert("c" in a and "b" in a and "a" in a)
        assert(a["a"]==a["A"])
        self.test.debug("init")
        b=CaselessDict(zip(["a","b","A"],[1,2,3]))
        assert(b["a"]==b["A"])
        assert(b["a"]==3)
        assert(len(b.keys())==2)
        #test pop

        self.test.debug("pop")
        q=b.pop("a")
        assert(q==3)
        assert(len(b.keys())==1)
        assert(b["b"]==2)
        #test pop with different case
        b["a"]=2
        assert(len(b.keys())==2)
        b.pop("A")
        assert(len(b.keys())==1)
        #test pop with invalid key
        b["a"]=2
        assert(len(b.keys())==2)
        try:
            b.pop("q")
            assert(False)
        except:
            assert(True)
        assert(len(b.keys())==2)

    def test_copy(self):
        self.test.debug("copy")
        a=CaselessDict({"a":1,"b":2})
        b=a.copy()
        assert(a==b)
        assert(a["a"]==b["a"])
        assert(a["a"]!=b["b"])
        assert(a["b"]==b["b"])

    def test_case(self):
        self.test.debug("case match")
        a=CaselessDict({"a":1,"A":2,"b":3})
        assert(a["a"]==a["A"])
        assert(a["a"]==2)
        assert(a["A"]==2)
        assert(not a["b"] == 2)
        assert(not a["b"] == a["a"])


        
    def alltest(self):
        self.test_null()
        self.test_init()
        self.test_case()
        self.test_getset()
        self.test_addondict()
        self.test_copy()
        pass
    
    
    
if __name__ == "__main__":
    t=test_CaselessDict()
    t.alltest()
        



"""This module calculates the Planck function in various forms, usually between a lower and 
upper limit and for a single temperature. Two functions use weave.blitz code to 
speed up the functions.

"""

#planck module
import numpy as np
import dwell.constants as const


def planck_wavelength(wavelength, temp):
    """Calculate Black Body radiance at a single wavelength or wavelengths.
    
    :param wavelength: wavelengths, microns
    :type wavelength: numpy array

    :param temp: of temperatures, kelvin
    :type temp: numpy array, or float

    :returns: Planck function in units of W/m^2/ster/micron 
              calculated at wavelength and temperature
    :rtype: numpy array
    """
    
    metres_to_microns = 1.e6
    c1 = 2*const.h*const.c*const.c    * metres_to_microns**4     #1.911e8
    c2 = (const.h*const.c/const.k_B) * metres_to_microns     #1.439e4
    Blambda = c1 * (wavelength)**(-5)/(np.exp(c2/(wavelength*temp))-1)
    
    return Blambda

    
def planck_wavenumber(wavenumber, temp):
    """Calculate Black Body radiance at a given wavenumber or set of wavenumbers
    
    :param wavenumber: wavenumber, cm-1
    :type wavenumber: numpy array

    :param temp: of temperatures, kelvin
    :type temp: numpy array or float

    :returns: Planck function in units of W/m^2/ster/cm-1 
              calculated at wavelength and temperature
    :rtype: numpy array
    """
    
    metres_to_centimetres = 1e2
    
    c1 = 2*const.h*const.c*const.c * metres_to_centimetres**4 #1.e-8
    c2 = const.h*const.c/(const.k_B) * metres_to_centimetres # 1.439
    
    Bnu = c1 * (wavenumber**3) / (np.exp(wavenumber*c2/temp) - 1)
    
    return Bnu

def cfst4_wavenumber(wl, wh, *args, **kwargs):
    """Calculate the integrated Planck function between two wavenumbers
    
    :param wl: low wavenumber, cm-1
    :type wl: numpy array
    
    :param wh: high wavenumber, cm-1
    :type wh: numpy array

    :param temp: of temperatures, kelvin
    :type temp: float

    :returns: Planck function in units of W/m^2/ster
              at temperature integrated between wl and wh
    :rtype: numpy array
    """
    return cfst4_wavelength(10000./wh,10000./wl, *args, **kwargs)

def cfst4_wavelength(wl, wh, temperature, diff_planck=None, diff_planck_2=None):
    """Calculate the integrated Planck function between two wavelengths
    
    :param wl: low wavelength, micron
    :type wl: numpy array
    
    :param wh: high wavelength, micron
    :type wh: numpy array

    :param temp: of temperatures, kelvin
    :type temp: float

    :returns: Planck function in units of W/m^2/ster
              at temperature integrated between wl and wh
    :rtype: numpy array

    """
    c=0.153989733
    c2 = 1.43883e4
    sigma = const.sigma
    
    wn = [wl,wh]
    fac = 0.0000001
    f=[0.0,0.0]
    for i in range(2):
        f[i]=0.0
        v = c2/(wn[i]*temperature)
        for m in range(1,1000):
            wv=float(m)*v
            ff=(float(m)**(-4))*(np.exp(-wv))*(((wv+3.0)*wv+6.0)*wv+6.0)
            f[i]=f[i]+ff
            if f[i] == 0.0:
                continue
            if ff/f[i] <= fac :
                break
            pass
        pass
    fst4 = c*(f[1]-f[0])*sigma*temperature**4
        
    
    return fst4

def cfst4_wavelength_weave(wl, wh, temperature, diff_planck=None, diff_planck_2=None):
    """Calculate the integrated Planck function between two wavenumbers. Uses 
    blitz.weave code to speed up the calculation.
    
    :param wl: low wavelength, micron
    :type wl: numpy array
    
    :param wh: high wavelength, micron
    :type wh: numpy array

    :param temp: of temperatures, kelvin
    :type temp: float

    :returns: Planck function in units of W/m^2/ster
              at temperature integrated between wl and wh
    :rtype: numpy array
    """
    c=0.153989733
    c2 = 1.43883e4
    sigma = const.sigma
    
    wn = np.array([wl,wh])
    fac = 0.0000001
    f=np.array([0.0,0.0])
    code="""
    double v,ff,wv,fst4;
    for (int i=0;i<2;i++){
        f(i)=0.0;
        v = c2/(wn(i)*temperature);
        ff=1e6;
        for (int m=1;m<1000;m++){
            wv=m*v;
            ff=(exp(-4.0*log(m)))*(exp(-wv))*(((wv+3.0)*wv+6.0)*wv+6.0);
            f(i)=f(i)+ff;
            if (f(i) == 0.0)  continue;
            if (ff/f(i) <= fac) break;
      }
      }

    fst4 = c*(f(1)-f(0))*sigma*exp(4*log(temperature));
    return_val = fst4;
    """
    from scipy.weave import converters
    import scipy.weave as weave
    err = weave.inline(code,['f','temperature', 'wn', 'fac','sigma','c','c2'],
                       type_converters=converters.blitz,compiler='gcc')
    
    return err

def cfst4_wavelength_weave_loop(wl, wh, temperature, diff_planck=None, diff_planck_2=None):
    """Calculate the integrated Planck function between two wavelengths. Uses 
    blitz.weave code to speed up the calculation
    :param wl: low wavelength, micron
    :type wl: numpy array
    
    :param wh: high wavelength, micron
    :type wh: numpy array

    :param temp: of temperatures, kelvin
    :type temp: float

    :returns: Planck function in units of W/m^2/ster
              at temperature integrated between wl and wh
    :rtype: numpy array

     
    """
    c=0.153989733
    c2 = 1.43883e4
    sigma = const.sigma
    
    wn = np.array([wl,wh])
    fac = 0.0000001
    f=np.array([0.0,0.0])
    len_t = temperature.size
    res=np.zeros(len_t, d_type=temperature.d_type)
    code="""
    double v,ff,wv,fst4;
    for(int l=0;l<len_t;l++){
    for (int i=0;i<2;i++){
        f(i)=0.0;
        v = c2/(wn(i)*temperature[l]);
        ff=1e6;
        for (int m=1;m<1000;m++){
            wv=m*v;
            ff=(exp(-4.0*log(m)))*(exp(-wv))*(((wv+3.0)*wv+6.0)*wv+6.0);
            f(i)=f(i)+ff;
            if (f(i) == 0.0)  continue;
            if (ff/f(i) <= fac) break;
      }
      }
    fst4 = c*(f(1)-f(0))*sigma*exp(4*log(temperature[i]));
    res[i]=fst4      
      }
      
    return_val = fst4;
    """
    from scipy.weave import converters
    import scipy.weave as weave
    err = weave.inline(code,['f','temperature', 'wn', 'fac','sigma','c','c2'],
                       type_converters=converters.blitz,compiler='gcc')
    
    return res


def st4(temp):
    """Calculates the integrated Black body function from zero to infinity,
    (wavenumber or wavelength),   sigma*temp^4.
    
    :param temp: temperature, K
    :type temp: numpy array, or float

    :returns: Planck function in units of W/m^2
              at temperature
    :rtype: same as input
    """
    
    bb = const.sigma * temp**4
    return bb

if __name__=="__main__":
    import dwell.testing.test_planck as pp
    c = pp.test_planck()
    c.allt()



"""Interface to logging to allow color logs to
terminal, or non-color to files"""
# termcolor
# if we have the termcolor package, then colorize some text.
import six
try:
    from termcolor import cprint, colored
    def colorize(color, on_color=None):
        return lambda text: colored(text,color=color, on_color=on_color)

    text_message = colorize('blue')
    text_info    = colorize("green")
    text_warning = colorize('magenta')
    text_error = colorize('red')
    text_critical = colorize('red')
    text_note = colorize('grey')
except:

    def nocolor(text, *args, **kwargs):
        return "nocolor: "+text

    def colored(text, color=None, on_color=None, attrs=None):
        return text
    pass
    text_message = nocolor
    text_warning = nocolor
    text_error = nocolor
    text_critical = nocolor
    text_info = nocolor
    text_note = nocolor

from logging import *
import sys

def handler(level=None, form=None, filename=None, stream=None, datefmt=None, use_color=True):
    """Setup for a handler:
        Level sets the minimum level noted in this log
        form sets the format of the log message
        filename is the filename for a file logger
        stream is the stream object for a stream logger (also colorizes)
        use_color overrides the stream color usage if False.
    """
    
    if datefmt is None:
        datefmt = "%Y/%m/%d %H:%M:%S "
    #configure the lowest level to print
    if level is None:
        level = DEBUG
    #configure the format of the output
    if form is None:
        form = "%(levelname)-8s %(asctime)s %(message)s"
    #set the output
    
    handler = None
    
    if stream is not None:
        handler = StreamHandler()
        class MyFormatter(Formatter):
            def __init__(self, fmt, datefmt=None):
                Formatter.__init__(self, fmt, datefmt)
            def colorize(self,str, levelname):
                """Attempts to colorize the string based on the
                warning level of the log message."""
                colors = {  "DEBUG": text_message,
                             "INFO": text_info,
                             "WARNING": text_warning,
                             "ERROR": text_error,
                             "CRITICAL": text_critical}
                if levelname in colors:
                    return colors[levelname](str)
                else:
                    return str#terminal.note(str)

            def format(self,record):
                str = Formatter.format(self, record)
                return self.colorize(str, record.levelname)
        if use_color:
            lform = MyFormatter(form, datefmt=datefmt)
        else:
            lform = Formatter(form,datefmt=datefmt)
        handler.setFormatter(lform)
        handler.setLevel(level)
    
    if filename is not None:
        handler = FileHandler(filename)
        lform = Formatter(form, datefmt=datefmt)
        handler.setFormatter(lform)
        handler.setLevel(level)

    return handler


def ProgramLogger(name, destinations):
    """Function to define multiple loggers in one step, and name them.
    destinations is a dictionary of dictionaries, where each dictionary
    contains level, form, filename OR stream, and use_color as required
    by handler above.
    This function iterates through the dictionary defining the handlers
    and adding them to the logger before returning."""

    log = getLogger(name)
    if len(log.handlers):
        return log
    log.setLevel(DEBUG)
    for key, dest in six.iteritems(destinations):
        handle = handler(level=dest.get("level", DEBUG),
                         form=dest.get("form", None),
                         filename=dest.get("filename", None),
                         stream=dest.get("stream", None),
                         use_color=dest.get("use_color", True))
        log.addHandler(handle)

    return log

def log_level(lookfor):
    """Tries to find a suitable log level based on input string. 
    Input can be a level (0=critical, 4=debug), or a substring of
    debug,info, warning, error, or critical.
    """
    levels={0:CRITICAL,
            1:ERROR,
            2:WARNING,
            3:INFO,
            4:DEBUG,
            "critical":CRITICAL,
            "error":ERROR,
            "warning":WARNING,
            "info":INFO,
            "debug":DEBUG
                }
    if lookfor.lower() in levels:
        return levels[lookfor.lower()],lookfor.lower()
    else:
        for k,v in six.iteritems(levels):
            l=len(str(k)) if len(str(k)) < len(lookfor) else len(lookfor)
            if str(k)[0:l].lower() == lookfor.lower():
                return v,k
    
    raise IndexError
    
def default_logger(name, logfile=None,
                    file_verbosity = "WARNING",
                    verbosity="CRITICAL"):
    """Sets up a simple logger in a file and the stderr, with standard error levels.
    This is intended for use with input options corresponding to the input options"""
    
    log_error=""
    try:
        level,log_level_name = log_level(verbosity)
    except IndexError:
        log_error =  log_error + "verbosity level invalid({0}, should be 0-4)".format(verbosity)
        level = CRITICAL
        log_level_name="critical"
    try:
        file_level, file_level_name = log_level(file_verbosity)
    except:
        log_error =  log_error + "file_verbosity level invalid({0}, should be 0-4)".format(file_verbosity)
        file_level = ERROR
        file_level_name="error"
    #stdout
    dest = {}
    stderr_h = dict(level=level, stream=sys.stderr, use_color=True)
    dest["stderr"]=stderr_h
    if logfile is not None:
        file_err = dict(level=file_level, filename=logfile, use_color=True)
        dest["logfile"]=file_err

    log = ProgramLogger(name, dest)
    log.setLevel(level if level < file_level else file_level)

    if log_error!="":
        log.error(log_error)
    log.info("log screenlevel is set to {0}".format(log_level_name))
    log.info("log filelevel is set to {0}".format(file_level_name))
    
    return log

def command_line(parser, progname, default_verbosity="CRITICAL", default_file_verbosity="CRITICAL"):
    """Adds command line options to the supplied parser instance"""
    parser.add_argument("--logfile",type=str, default="{0}.log".format(progname))
    parser.add_argument("--verbosity","-v",type=str, default=default_verbosity)
    parser.add_argument("--file_verbosity","-fv",type=str, default=default_file_verbosity)
    


"""Terminal utilities"""

# termcolor
# if we have the termcolor package, then colorize some text.
try:
    from termcolor import cprint, colored
    def colorize(color, on_color=None):
        return lambda text: colored(text,color=color, on_color=on_color)

    message = colorize('blue')
    info    = colorize("green")
    warning = colorize('magenta')
    error = colorize('red')
    critical = colorize('red')
    note = colorize('grey')
except:

    def nocolor(text, *args, **kwargs):
        return text

    def colored(text, color=None, on_color=None, attrs=None):
        return text
    pass
    message = nocolor
    warning = nocolor
    error = nocolor
    critical = nocolor
    info = nocolor
    note = nocolor


try:
    from argh import ArghParser, arg
    import log as mylog
    import sys
    def argh_main(progname, commands,default_verbosity=None, default_file_verbosity=None):
        default_verbosity=default_verbosity or "CRITICAL"
        default_file_verbosity=default_file_verbosity or "CRITICAL"
        parser = ArghParser()
        parser.add_argument("--very_verbose", action="store_true")
        parser.add_argument("--logfile",type=str, default="{0}.log".format(progname))
        parser.add_argument("--verbosity","-v",type=str, default=default_verbosity)
        parser.add_argument("--file_verbosity","-fv",type=str, default=default_file_verbosity)
        #mylog.command_line(parser, progname, default_verbosity="CRITICAL", default_file_verbosity="INFO")

        parser.add_commands(commands)
        args = parser.parse_args()
    
        log = mylog.default_logger(progname,
                                    logfile=args.logfile,
                                    verbosity=args.verbosity, 
                                    file_verbosity=args.file_verbosity)
    
        try:
            parser.dispatch()
        except KeyboardInterrupt as e:
            log.info("Exiting on Keyboard Interrupt")
            print("Exiting on Keyboard Interrupt")
        except Exception as e:
            print("ERROR: {0}".format(e))
            log.error(e.message)
            if args.very_verbose:
                raise
            else:
                sys.exit(1)
except:
    def argh_main(progname, commands):
        print ("Import of argh module failed")
        import sys
        sys.exit(1)



from boolparser import BoolParser, EvaluateVariable


class ncar(object):
    def __init__(self,var):
        self.var = var
        return
    def __ge__(self, num):
        return self.var>=num
    def __le__(self, num):
        return self.var<=num
    def __gt__(self, num):
        return self.var>num
    def __lt__(self, num):
        return self.var<num
    
def define_class(nc):
    class ev_nc(EvaluateVariable):
       def eval(self):
            if self.value in nc.variables:
                return ncar(nc.variables[self.value][:])
            elif self.value in nc.dimensions:
                return ncar(nc.dimensions[self.value][:])
            else:
                raise NameError("Not a variable: {0}".format(self.value))
                return self.value
    return ev_nc


def boolfilter(nc, condition):
    nceval = define_class(nc)
    bp = BoolParser(nceval)
    return bp.parse(condition)
    
